import requests as _requests

from .. import package_consts as _Const
from ..connection.repeated_call import request_or_repeat as _request_or_repeat

_KEY_FILE_URL = "json_data_file_url"
_KEY_IS_LOADABLE = "is_json_data_loadable"


def get_meta_by_id(snapshot_id: int) -> dict:
    response = _request_or_repeat(
        _requests.get, url=f"{_Const.BACKEND_SNAPSHOT_URL}/{snapshot_id}/head"
    )
    response.raise_for_status()
    return response.json()


def _get_body_by_url_or_id(
    *,
    is_loadable: None | bool = None,
    body_file_url: None | str = None,
    snapshot_id: None | int = None,
):
    if body_file_url is None and snapshot_id is None:
        raise ValueError(
            "Cant get snapshot body without args. Please give file url of snapshot data or snapshot_id"
        )
    if is_loadable == False:
        # why bother downloading if it will be blank / corrupted / etc. ?
        # or should exception be raised?
        return {}
    if body_file_url is None:
        # maybe old snapshot, so ask for preview, expecting to get full snapshot body
        body_file_url = f"{_Const.BACKEND_SNAPSHOT_URL}/{snapshot_id}/preview"
    response = _request_or_repeat(_requests.get, url=body_file_url)
    response.raise_for_status()
    return response.json()


def get_body_by_id(snapshot_id: int):
    # get meta
    meta = get_meta_by_id(snapshot_id)
    # now body
    return _get_body_by_url_or_id(
        is_loadable=meta.get(_KEY_IS_LOADABLE, None),
        body_file_url=meta.get(_KEY_FILE_URL, None),
        snapshot_id=snapshot_id,
    )


def _raise_if_not_implemented_filter(
    *,
    by_is_verified: None | bool = None,
    by_validation_statuses: None | list[str] = None,
    **_,
) -> None:
    if by_is_verified is not None:
        raise NotImplemented("Sorry, filter is_verified not implemented")
    elif by_validation_statuses is not None:
        raise NotImplemented("Sorry, filter validation_statuses not implemented")


def _args_to_filter_params(
    *,
    by_challenge_id: None | int = None,
    by_user_id: None | int = None,
    by_is_verified: None | bool = None,
    by_validation_statuses: None | list[str] = None,
    by_is_from_robot: None | bool = None,
    **_,
) -> dict:
    return {
        "challenge_id": by_challenge_id,
        "user_id": by_user_id,
        "is_verified": by_is_verified,
        "validation_statuses": by_validation_statuses,
        "is_from_code_run": by_is_from_robot,
    }


def get_latest_meta(
    *,
    by_challenge_id: None | int = None,
    by_user_id: None | int = None,
    by_is_verified: None | bool = None,
    by_validation_statuses: None | list[str] = None,
    by_is_from_robot: None | bool = None,
) -> dict:
    # check, is each filter supported / implemented.
    # at this point locals() will have dict of supplied args and their values
    _raise_if_not_implemented_filter(**locals())
    response = _request_or_repeat(
        _requests.get,
        url=f"{_Const.BACKEND_SNAPSHOT_URL}/latest/",
        params=_args_to_filter_params(**locals()),
    )
    response.raise_for_status()
    return response.json()


def get_latest_body(
    *,
    by_challenge_id: None | int = None,
    by_user_id: None | int = None,
    by_is_verified: None | bool = None,
    by_validation_statuses: None | list[str] = None,
    by_is_from_robot: None | bool = None,
):
    # at this point locals() will have dict of supplied args and their values
    meta = get_latest_meta(**locals())
    return _get_body_by_url_or_id(
        is_loadable=meta.get(_KEY_IS_LOADABLE, None),
        body_file_url=meta.get(_KEY_FILE_URL, None),
        snapshot_id=meta.get("id"),
    )



from requests import Response
from requests.exceptions import ConnectionError
from time import sleep

from .. import package_consts as _Const


def request_or_repeat(function: callable, *args, **kwargs) -> Response:
    """Makes a request (call provided function with args and kwargs).
    * If it fails because of ConnectionError, retries;
    * If it fails because status_code >= 500, retries;
    retry count and sleep times are specified in package_consts.py

    :param function: function to call; expected to get response object from it.
    :param *args: what args to pass to function.
    :param **kwargs: what kwargs to pass to function.
    :return: `Response` or raises RuntimeError after all failures
    """

    failures_count = 0
    sleep_time = _Const.REQUEST_INITIAL_SLEEP_AFTER_FAILURE_IN_S

    def _handle_failure():
        nonlocal failures_count, sleep_time
        sleep_time = (
            sleep_time**_Const.REQUEST_SLEEP_INCREASE_POWER
            if failures_count > 0
            else sleep_time
        )
        failures_count = failures_count + 1

    while failures_count < _Const.REQUEST_DROP_ON_FAILURE_COUNT:
        if failures_count > 0:
            sleep(sleep_time)
        try:
            response: Response = function(*args, **kwargs)
            if response.status_code >= 500:
                # server tripped, not we (internal server error)
                raise RuntimeError
            else:
                return response
        except ConnectionError:
            _handle_failure()
        except RuntimeError:
            _handle_failure()

    raise RuntimeError(
        f"[internal] Could not make connection for {_Const.REQUEST_DROP_ON_FAILURE_COUNT} times."
    )



# Constants used by package

REQUEST_DROP_ON_FAILURE_COUNT = 4
"How many `didnt connect` / `status >= 500` should occur sequentially before giving-up-on-request"

REQUEST_INITIAL_SLEEP_AFTER_FAILURE_IN_S = 3
"How many seconds to sleep after first request failure"

REQUEST_SLEEP_INCREASE_POWER = 2
"sleep time will be increased by this power. time ^ power"

BACKEND_URL = "https://data-platform-backend-4ddpl.ondigitalocean.app"
"url of server that stores snapshots, challenges, code runs. do not include ending slash"

BACKEND_SNAPSHOT_URL = f"{BACKEND_URL}/snapshots"
"backend url root for snapshots. it might be prepended in actual usage"
